/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2003-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.mathExpr;


import de.jtem.mathExpr.evaluator.Context;
import de.jtem.mathExpr.evaluator.Evaluator;
import de.jtem.mathExpr.evaluator.Type;
import de.jtem.mathExpr.parser.Expression;
import de.jtem.mathExpr.parser.Symbol;
import de.jtem.mathExpr.parser.UnaryOperation;
//for doc-links:

/**
 * A <code>Function</code> object represents a mathematical function with a
 * name, one or more parameters and a function definition. The function
 * definition is an {@link de.jtem.mathExpr.parser.Expression}.
 * <p>
 * A <code>Function</code> object is used for the <code>Evaluator</code> to be
 * able to evaluate {@link de.jtem.mathExpr.parser.FunctionCall}
 * <code>Expressions</code>.</p>
 * <p>
 * Before evaluating a <code>FunctionCall</code> in the
 * expression-tree-hirarchy, the appropriate <code>Function</code> must be
 * defined or added to the {@link DefinitionModel}, e.g.:
 * <br><br><code>
 * {@link de.jtem.mathExpr.evaluator.Type} defaultType=
 * {@link de.jtem.mathExpr.evaluator.realEvaluator.RealType#TYPE};<br>
 * {@link ExpressionConfiguration} config=new
 * ExpressionConfiguration(defaultType);<br>
 * //define the function f(x,y)=x+y<br>
 * config.{@link ExpressionConfiguration#defineFunction(String,String[],String)
 * defineFunction}("f", new String[] {"x","y"}, "x+y");<br>
 * config.{@link ExpressionConfiguration#setExpression(String) 
 * setExpression}("2*f(3,4)")</code>
 * <br><br>
 * All {@link PredefinedFunction PredefinedFunctions} were added to the
 * <code>DefinitionModel</code> by that constructor.
 * </p>
 * <p>
 * In opposite to {@link de.jtem.mathExpr.evaluator.Variable Variables}, a
 * <code>Function</code>
 * <b>must</b> be already defined before parsing a <code>String</code>
 * containing the appropriate <code>FunctionCall</code>, otherwise the
 * <code>Parser</code> can't differ between a <code>Variable</code> in
 * an implicit multiplication or a <code>Function</code> - it looks identical,
 * for example:<br>
 * <br>
 * You want to parse the <code>Expression</code> "f(2+2)".<br>
 * If the parsed identifier "f" is defined as a <code>Function</code>, the
 * <code>Parser</code> (respectively its {@link ExpressionConfiguration}) will
 * create a <code>FunctionCall</code>.<br>
 * But if "f" is defined as a <code>Variable</code> (or not at all), the <code>
 * Parser</code> (respectively its {@link ExpressionConfiguration}) will create
 * a {@link de.jtem.mathExpr.parser.BinaryOperation} "f*(2+2)".
 * </p>
 *
 * @see de.jtem.mathExpr.evaluator.Type#getEvaluator(UnaryOperation,Evaluator)
 */
public abstract class Function
{
  /**
   * This function's name.
   */
  protected String name;
  /**
   * This function's parameters.
   */
  protected String[] parameter;
  /**
   * Creates a new <code>Function</code> with the specified <code>name</code>
   * and <code>parameters</code>.
   * 
   * @param name the function's name.
   * @param p an array of the function's parameters.
   */
  public Function(String name, String[] p) 
  {
	  this.name=name;
	  parameter=p;
  }
  /**
   * Returns this' function name. The name of a <code>Function</code> will be
   * defined by any constructor. A <code>Function</code> can't be renamed.
   * 
   * @return this' function name.
   */
  public String getName() 
  {
    return name;
  }
  /**
   * Returns the number of parameters this <code>Function</code> depends on.
   * 
   * @return the number of parameters of this <code>Function</code>.
   */
  public int getParameterCount() 
  {
    return parameter.length;
  }
  /**
   * Sets the parameter names of this <code>Function</code> to the
   * <code>Strings</code> in the specified array.
   * 
   * @param newParameters an array of the new parameter names.
   */
  public void setParameters(String[] newParameters)
  {
    parameter=newParameters;
  }
  /**
   * Returns all parameter names.
   * 
   * @return all parameters names.
   */
  public String[] getParameters() 
  {
    return parameter;
  }
  /**
   * Returns the index' parameter name.
   * 
   * @param index the index of the parameter which name is to be returned.
   * @return the index' parameter name.
   */
  public String getParameter(int index) 
  {
    return parameter[index];
  }
  /**
   * Returns the index of the parameter <code>param</code> or <code>-1</code>
   * if this <code>Function</code> does not have that parameter.
   * 
   * @return the index of the parameter <code>param</code> or <code>-1</code>
   * if this <code>Function</code> does not have that parameter.
   * @param param the parameter which index is in demand.
   */
  public int getParameterIndex(String param) 
  {
    for(int i=0; i<parameter.length; i++)
      if(parameter[i].equals(param))
        return i;
    return -1;
  }
  /**
   * Return the <code>Evaluator</code> for this <code>Function</code>. If this
   * Function is a {@link PredefinedFunction}, this method will call the method
   * {@link de.jtem.mathExpr.evaluator.Type#getEvaluator(PredefinedFunction,Evaluator)}
   * on that Type the parameter evaluator returns.
   * 
   * @return the <code>Evaluator</code> for this <code>Function</code>.
   * @param defaultType the <code>Type</code>, <code>Symbols</code> will be
   * evaluated to.
   * @param context the <code>Context</code>, <code>Evaluators</code> for
   * <code>Symbols</code> will be get from.
   */
  public abstract Evaluator getCallEvaluator(Type defaultType, Context context);
  /**
   * Returns the derivative function of this Function with respect to the
   * paramIndex' parameter. 
   * 
   * @see de.jtem.mathExpr.parser.Expression#differentiate(Symbol)
   * @return a new <code>Function</code> that is the derivative of this with
   * respect to the paramIndex' parameter
   */
  public abstract Function getDerivative(int paramIndex);
  /**
   * Returns this function dissolved. Therefore the parameters in the
   * function definition will be replaced by the corresponding param.
   *  
   * @param param
   */
  public abstract Expression insertArguments(Expression param);//TODO: param musst be an Expression Array (or List???) (for more than one parameter)
}
