/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2003-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.mathExpr;

import de.jtem.mathExpr.parser.List;

/**
 * Provides static methods to check if specified de.jtem.mathExpr.parser.List objects
 * are in a certain mathematical form (e.g.: a rectangular Matrix).
 *
 * @see de.jtem.mathExpr.parser.List
 * @author marcel
 */
public class ListFormChecker
{
  /**
   * Checks if the specified List represents a mathematical Matrix.
   * Returns true if the List object represents a rectangular Matrix,
   * that means:
   * <ul>
   * <li>all entries are (Sub-)Lists AND</li>
   * <li>all (Sub-)Lists have the same number of entries AND</li>
   * <li>these entries are no (Sub-Sub-)Lists.</li>
   * </ul>
   * <br>
   * Returns false if the List is not conform to at least one of these three
   * items.
   */
  static public boolean representsMatrix(List list)
  {
    int listSize=list.getEntryCount();
    List entry;
    int expectedSubListSize=0;
    for(int i=0; i<listSize; i++)
    {
      //---- first item ----
      try
      {
        //if(AbstractEvaluatorFactory.CAST_DEBUG)
        //  AbstractEvaluatorFactory.castLog(
        //      "DoubleEvaluatorFactory.getUnaryOperationEvaluator");
        entry=(List)list.getEntry(i);
      }
      catch(ClassCastException ex)
      { 
        return false;//one entry is no list
      }
      //---- second item ----
      if(i==0)
        expectedSubListSize=entry.getEntryCount();
      else
      {
        int subListSize=entry.getEntryCount();
        if(subListSize!=expectedSubListSize)
          return false;//different sub-list sizes
      }
      //---- third item ----
      if(!representsVector(entry))
        return false;//sub-list entry is list
    }
    return true;
  }
  /**
   * Returns true if the specified List object represents a mathematical Vector,
   * that means all entries are no (Sub-)Lists, otherwise returns false.
   */
  static public boolean representsVector(List list)
  {
    int listSize=list.getEntryCount();
    for(int i=0; i<listSize; i++)
    {
      if(list.getEntry(i) instanceof List)
        return false;
    }
    return true;
  }
  
}
