/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2003-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.mathExpr.evaluator;

import de.jtem.mathExpr.evaluator.complexEvaluator.ComplexConstant;
import de.jtem.mathExpr.evaluator.complexEvaluator.ComplexType;
import de.jtem.mathExpr.evaluator.realEvaluator.Real;
import de.jtem.mathExpr.evaluator.realEvaluator.RealConstant;
import de.jtem.mathExpr.evaluator.realEvaluator.RealType;
import de.jtem.mathExpr.parser.Symbol;
import de.jtem.mfc.field.Complex;

/**
 * Represents a mathematical number, which value never changes.
 * {@link #PI} and {@link #E} are certain <code>Constant</code> objects
 * representing the appropriate mathematical constant.
 * <br><br>
 * A <code>Constant</code> will be created by the method {@link
 * Type#getEvaluator(Symbol,Context)} to return the Evaluator - that
 * created <code>Constant</code> - for a {@link Symbol} representing
 * a constant number. Such <code>Constants</code> are instances of
 * {@link RealConstant} or {@link ComplexConstant} in any case because they
 * are part of an evaluator-tree.<br>
 * <br>
 * If a <code>Symbol</code> represents no constant
 * number but nevertheless a <code>Constant</code> like "pi" or "e", the
 * method {@link #getConstant(Symbol,Type)} is used to return that
 * <code>Constant</code> object.
 */
public class Constant implements Evaluator
{
	/**
	 * Represents the mathematical constant pi=3.1415....
	 * Its {@link Type} is {@link RealType#TYPE}.
	 * 
	 * @see Math#PI
	 */
  static public final Constant PI=
      new RealConstant("pi", new Real(Math.PI), RealType.TYPE);
  /**
   * Represents the mathematical constant e=2.7182....
   * Its {@link Type} is {@link RealType#TYPE}.
	 * 
	 * @see Math#E
   */
  static public final Constant E=
      new RealConstant("e", new Real(Math.E), RealType.TYPE);
  /**
   * Represents the complex constant i=sqrt(-1).
   * Its {@link Type} is {@link ComplexType#TYPE}.
   * 
   * @see Complex
   */
  static public final Constant I=
      new ComplexConstant("i", new Complex(0,1), ComplexType.TYPE);
  /**
   * An array of all <code>static final Constant</code> fields of this class.
   */
  static public final Constant[] CONSTANTS={PI, E, I};
  /**
   * The name of this <code>Constant</code>.
   */
  final String name;
  /**
   * The value of this <code>Constant</code>.
   */
  final Object value;
  /**
   * The <code>Type</code> of this <code>Constant</code>.
   */
  final Type type;
  /**
   * Creates a new <code>Constant</code> with the name <code>n</code>,
   * the value <code>val</code> and the <code>Type t</code>. The
   * <code>Type</code> is necessary for the correct evaluation of this
   * <code>Constant</code> in relation to other {@link Evaluator Evaluators}
   * with different evaluation <code>Types</code>..
   *  
   * @param n the name for the new <code>Constant</code>.
   * @param val the value for the new <code>Constant</code>.
   * @param t the <code>Type</code> of the value of the new
   * <code>Constant</code>.
   */
  public Constant(String n, Object val, Type t)
  {
    name=n;
    value=val;
    type=t;
  }
	/**
	 * Either returns the <code>Constant</code> represented by the specified
	 * <code>Symbol</code> and <code>Type</code> if contained by the
	 * {@link #CONSTANTS} array, or returns <code>null</code> if an appropriate
	 * <code>Constant</code> does not exist in that array.
	 * For example, this method will be used for the evaluation of
	 * <code>Symbols</code> to decide if that <code>Symbol</code> is representing
	 * a {@link Variable} or a <code>Constant</code>.
	 * 
	 * @see Type#getEvaluator(Symbol,Context)
	 * @return the <code>Constant</code> represented by the specified
	 * <code>Symbol</code> and <code>Type</code> if contained by the
	 * {@link #CONSTANTS} array, otherwise <code>null</code>.
	 * @param s the <code>Symbol</code> representing the <code>Constant</code>
	 * to be searched for in the {@link #CONSTANTS} array.
	 * @param t the <code>Type</code> the <code>Constant</code>
	 * must be of.
	 */
	static public Constant getConstant(Symbol s, Type t)
	{
		String name=s.toString();
		for(int i=0; i<CONSTANTS.length; i++)
		{
			if(name.equalsIgnoreCase(CONSTANTS[i].toString()) &&
					t==CONSTANTS[i].getReturnType())
				return CONSTANTS[i];
		}
		return null;
	}
	/**
	 * Returns <code>true</code> if the specified <code>Symbol</code> represents
	 * a <code>Constant</code> contained by the {@link #CONSTANTS} array,
	 * <code>false</code> otherwise.
	 * 
	 * @return <code>true</code> if the specified <code>Symbol</code> represents
	 * a <code>Constant</code> contained by the {@link #CONSTANTS} array,
	 * <code>false</code> otherwise.
	 * @param s the <code>Symbol</code> representing the <code>Constant</code>
	 * to be searched for in the {@link #CONSTANTS} array.
	 */
	static public boolean isConstant(Symbol s)
	{
		String name=s.toString();
		for(int i=0; i<CONSTANTS.length; i++)
		{
			if(name.equalsIgnoreCase(CONSTANTS[i].toString()))
				return true;
		}
		return false;    
	}
  /**
   * Evaluates this <code>Constant</code> by returning its value.
   * 
   * @return the value of this <code>Constant</code> as an <code>Object</code>.
   */
  public Object evaluate()
  {
    return value;
  }
  /**
   * Returns the <code>Type</code> of this <code>Constant's</code> value.
   * 
   * @return the <code>Type</code> of this <code>Constant's</code> value.
   */
  public Type getReturnType()
  {
    return type;
  }
	/**
	 * Returns the name of this <code>Constant</code>.
	 * 
	 * @return the name of this <code>Constant</code>.
	 */
	public String toString()
	{
		return name;
	}
}
