/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2003-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.mathExpr.evaluator;

//for doc-links:
import de.jtem.blas.ComplexVector;
import de.jtem.blas.RealVector;
import de.jtem.mathExpr.ExpressionConfiguration;
import de.jtem.mathExpr.evaluator.complexEvaluator.AbstractComplexEvaluator;
import de.jtem.mathExpr.evaluator.complexVectorEvaluator.AbstractComplexVectorEvaluator;
import de.jtem.mathExpr.evaluator.realEvaluator.AbstractRealEvaluator;
import de.jtem.mathExpr.evaluator.realEvaluator.Real;
import de.jtem.mathExpr.evaluator.realVectorEvaluator.AbstractRealVectorEvaluator;
import de.jtem.mathExpr.parser.Expression;
import de.jtem.mathExpr.parser.javacc.Parser;

/**
 * The interface for all evaluators which can be created from an {@link
 * Expression}.
 * Similar to the {@link Parser} building up an expression-tree-hirarchy from
 * a given <code>String</code>, the method {@link
 * Expression#getEvaluator(Type,Context)}
 * of an {@link Expression} (representing that expression-tree-hirarchy) builds
 * up an evaluator-tree-hirarchy in a kind of recursion and returns it as an
 * <code>Evaluator</code> object.
 */
public interface Evaluator
{
  /**
   * Evaluate an {@link Expression} and return the result as an
   * <code>Object</code>.
   * <p>
   * <b>Attention:</b> The returned <code>Object</code> and the result
   * <code>Object</code> storing in a field by this <code>Evaluator</code>
   * are the same. If you need a new instance for the result, you have to assign
   * the returned value to a new one. If the type of the evaluation result is
   * known, you can also use the <code>get...Value(...)</code> method of your
   * current <code>Evaluator</code> to assign the evaluation result to a given
   * <code>Object</code> of the same type, for example:<br>
   * <pre>
   * //an {@link ExpressionConfiguration} with name "config" was defined and
   * //an {@link Expression} was already setted.
   * AbstractRealEvaluator e=(AbstractRealEvaluator)config.getEvaluator();
   * e.evaluate();
   * Real r=new Real();
   * e.getRealValue(r);
   * </pre>
   * 
   * @see AbstractRealEvaluator#getRealValue(Real)
   * @see AbstractComplexEvaluator#getComplexValue(Complex)
   * @see AbstractRealVectorEvaluator#getRealVectorValue(RealVector)
   * @see AbstractComplexVectorEvaluator#getComplexVectorValue(ComplexVector)
   * @return the evaluation result.
   */
  public Object evaluate();
  /**
   * Return the <code>Type</code> of the evaluated <code>Object</code>.
   * 
   * @return the <code>Type</code> of the evaluation result.
   */
  public Type getReturnType();
}
