/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2003-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.mathExpr.evaluator;


//for doc-links:
import de.jtem.mathExpr.DefinitionModel;
import de.jtem.mathExpr.ExpressionConfiguration;
import de.jtem.mathExpr.Function;
import de.jtem.mathExpr.evaluator.realEvaluator.Real;
import de.jtem.mathExpr.parser.Expression;
import de.jtem.mathExpr.parser.FunctionCall;
import de.jtem.mathExpr.parser.Symbol;

/**
 * The <code>Evaluator</code> evaluating any {@link FunctionCall}.
 * <p>
 * This class implements {@link Context} by resolving parameter names
 * and delegating other symbols and function calls to the global Context.
 * </p>
 */
public class FunctionCallEvaluator implements Evaluator, Context
{
  /**
   * The <code>Function</code> this <code>FunctionCallEvaluator</code>
   * deals with.
   */
  Function function;
  /**
   * The <code>Evaluator</code> for the function definition.
   */
  Evaluator functionEvaluator;
  /**
   * The <code>Evaluators</code> for the function parameters.
   */
  Evaluator[] pEvaluators;
  /**
   * After evaluation of all parameters (when {@link #evaluate()} is called
   * the first time), the evaluated values will be stored in this array.
   * That is to avoid repeated evaluations of the same parameter
   * (e.g.: <code>f(x)=x+x+x</code>).
   */
  private Object[] pValues;
  /**
   * The global {@link Variable} <code>Context}. This
   * <code>FunctionCallEvaluator</code> is a <code>Variable Context</code>
   * itself to ensure that the parameter evaluators were used for the
   * appropriate <code>Variable</code> in the function definition. This
   * <code>globalContext</code> is necessary to get even the
   * <code>Variables</code> which are not any parameters but defined by the
   * method {@link ExpressionConfiguration#defineVariable(String,Object,Type)}.
   * 
   * @see #getParameterEvaluator(int)
   */
  Context globalContext;
  /**
   * The <code>defaultType</code> specifies, how to deal with {@link Symbols}
   * in the expression-tree-hirarchy.
   */
  Type defaultType;
  /**
   * Creates a new <code>FunctionCallEvaluator</code> for the specified
   * <code>Function</code>.
   * <code>defaultType</code> is the <code>Type</code> from which the method
   * {@link Type#getEvaluator(FunctionCall,Context)} was called - in other
   * words: It specifies, how to deal with {@link Symbol Symbols} in the
   * expression-tree-hirarchy (e.g.: evaluate any <code>Symbol</code> as a
   * {@link de.jtem.mfc.field.Complex} or a {@link Real} or whatever).
   *
   * @see Function#getCallEvaluator(Type,Context)
   * @param f the <code>Function</code> which is to be evaluated.
   * @param p an array of all parameter <code>Expressions</code> of the
   * {@link FunctionCall}.
   * @param defaultType the <code>Type</code> any {@link Symbol} will be
   * evaluated to.
   * @param globalContext the <code>Context</code>, the <code>Evaluators<code>
   * for the <code>Symbols</code> in the function definition, which are no parameters,
   * are getting from.
   */
  public FunctionCallEvaluator(Function f, Expression[] p,
      Type defaultType, Context globalContext)
  {
  	//TODO: optimize for one parameter only
    this.defaultType=defaultType;
    this.globalContext=globalContext;
    function=f;
    pEvaluators=new Evaluator[p.length];
    for(int i=0; i<p.length; i++)
      pEvaluators[i]=p[i].getEvaluator(defaultType, globalContext);
    //no matter what kind of function (pre- or userdefined)
    functionEvaluator=f.getCallEvaluator(defaultType, this);
  }
  /**
   * Evaluates a {@link FunctionCall} and returns the value as an
   * <code>Object</code>. The <code>Type</code> of that <code>Object</code>
   * will be returned by the method {@link #getReturnType()}.
   * 
   * @return the value of the evaluated <code>FunctionCall</code>.
   */
  public Object evaluate()
  {
    if(pValues==null)
    {
      pValues=new Object[pEvaluators.length];
    }
    for(int i=0; i<pEvaluators.length; i++)
      pValues[i]=pEvaluators[i].evaluate();
    return functionEvaluator.evaluate();
  }
  /**
   * Returns the <code>Type</code> of the evaluated value.
   *
   * @see #evaluate()
   * @return the <code>Type</code> of the evaluated value.
   */
  public Type getReturnType()
  {
    return functionEvaluator.getReturnType();
  }
  /**
   * Returns the parameter evaluator or the global {@link Variable} regarding
   * to the specified <code>Symbol</code>.
   * The <code>Evaluator</code> for the function definition
   * gets this <code>FunctionCallEvaluator</code> as its Context to be able to decide 
   * between parameters and global variables in the function definition.
   * 
   * @return the parameter evaluator or the global {@link Variable} regarding
   * to the specified <code>Symbol</code>.
   * @param s the <code>Symbol</code> an <code>Evaluator</code> will be
   * returned for.
   */
  public Evaluator getEvaluator(Symbol s)
  {
    int index=function.getParameterIndex(s.toString());
    if(index==-1)
      return globalContext.getEvaluator(s);
    return getParameterEvaluator(index);
  }
  /**
   * Delegates the method {@link DefinitionModel#getFunction(FunctionCall)}
   * of the global {@link Context} to return the <code>Function</code> according
   * to the specified <code>FunctionCall</code> which is a sub-expression in
   * the function definition.
   * 
   * @return the <code>Function</code> according to the specified
   * <code>FunctionCall</code>.
   * @param fc the <code>FunctionCall</code> (a sub-expression in the
   * function definition) calling the <code>Function</code> to be return.
   */
  public Function getFunction(FunctionCall fc)
  {
    return globalContext.getFunction(fc);
  }
  /**
   * Returns a new <code>Evaluator</code> for the <code>index'</code> parameter.
   * 
   * @return a new <code>Evaluator</code> for the <code>index'</code> parameter.
   * @param index the index of the parameter which <code>Evaluater</code> is to
   * be returned.
   */
  public Evaluator getParameterEvaluator(int index)
  {
    return new ParameterEvaluator(index);
  }
  /**
   * A wrapper for an evaluated parameter.
   */
  private class ParameterEvaluator implements Evaluator
  {
    /**
     * The parameter index.
     */
    int index;
    /**
     * Creates a new <code>ParameterEvaluator</code> which evaluates the
     * index' <code>Object</code> in {@link FunctionCallEvaluator#pValues}
     * and which has the return {@link Type} of the index' parameter evaluator.
     */
    protected ParameterEvaluator(int i)
    {
      index=i;
    }
    /**
     * Does not evaluate but returns only the index' <code>Object</code> in
     * {@link FunctionCallEvaluator#pValue}.
     *
     * @see #index
     */
    public Object evaluate()
    {
      return pValues[index];
    }
    /**
     * Returns the ReturnType of the index' parameter evaluator.
     *
     * @see #index
     */
    public Type getReturnType()
    {
      return pEvaluators[index].getReturnType();
    }
  }
}
