/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2003-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.mathExpr.evaluator;

import de.jtem.mathExpr.PredefinedFunction;
import de.jtem.mathExpr.parser.BinaryOperation;
//for doc-links:
import de.jtem.mathExpr.parser.Expression;
import de.jtem.mathExpr.parser.FunctionCall;
import de.jtem.mathExpr.parser.List;
import de.jtem.mathExpr.parser.Symbol;
import de.jtem.mathExpr.parser.UnaryOperation;

/**
 * An interface for the variety of <code>Types</code> the result of
 * an evaluation can be.
 * This interface provides methods to get the appropriate {@link Evaluator} for 
 * a certain {@link Expression} in relation to the <code>Type</code> to what all
 * is evaluated, e.g.:<br><code>
 * Real * Real = Real<br>
 * Real * Complex = Complex<br>
 * Complex * Complex = Complex<br>
 * ComplexVector * Complex = ComplexVector<br>
 * ...</code><br>
 * <br>
 * The method {@link #getRepresentationClass()} returns the <code>Class</code>
 * of evaluation result this <code>Type</code> is representing.
 * 
 * @see Evaluator#getReturnType()
 */
public interface Type
{
	/**
	 * Return the <code>Evaluator</code> for the specified <code>Symbol</code>.
	 * An <code>Evaluator</code> for a <code>Symbol</code> will be either a
	 * {@link Variable} or a {@link Constant}.
	 * 
	 * @return the <code>Evaluator</code> for the specified <code>Symbol</code>.
	 * @param s the <code>Symbol</code> the <code>Evaluator</code>
	 * is to be returned for.
	 * @param context the <code>Context</code> the <code>Evaluator</code>
	 * will be got from.
	 * @see Context#getEvaluator(Symbol)
	 */
	Evaluator getEvaluator(Symbol s, Context context);
	/**
	 * Return the <code>Evaluator</code> for the specified
	 * <code>UnaryOperation</code>.
	 * 
	 * @return the <code>Evaluator</code> for the specified
	 * <code>UnaryOperation</code>.
	 * @param uo the <code>UnaryOperation</code> the <code>Evaluator</code>
	 * is to be returned for.
	 * @param e the <code>Evaluator</code> for the operand of the
	 * <code>UnaryOperation</code>.
	 */
	Evaluator getEvaluator(UnaryOperation uo, Evaluator e);
  /**
   * Return the <code>Evaluator</code> for the specified
   * <code>BinaryOperation</code>.
   * 
   * @return the <code>Evaluator</code> for the specified BinaryOperation.
   * @param bo the <code>BinaryOperation</code> the <code>Evaluator</code>
   * is to be returned for.
   * @param e1 the <code>Evaluator</code> for the first operand of the
   * <code>BinaryOperation</code>.
   * @param e2 the <code>Evaluator</code> for the second operand of the
   * <code>BinaryOperation</code>.
   */
  Evaluator getEvaluator(BinaryOperation bo, Evaluator e1, Evaluator e2);
	/**
	 * Return the <code>Evaluator</code> for the specified <code>List</code>.
	 * 
	 * @return the <code>Evaluator</code> for the specified <code>List</code>.
	 * @param l the <code>List</code> the <code>Evaluator</code>
	 * is to be returned for.
	 * @param entryEval an array of all <code>Evaluator</code> for the list
	 * entries.
	 * @see List#getEntries()
	 */
	Evaluator getEvaluator(List l, Evaluator[] entryEval);
  /**
   * Return the <code>Evaluator</code> for the specified
   * <code>FunctionCall</code>.
   * 
   * @return the <code>Evaluator</code> for the specified
   * <code>FunctionCall</code>.
   * @param fc the <code>FunctionCall</code> the <code>Evaluator</code>
   * is to be returned for.
   * @param context the evaluation <code>Context</code> the function's
   * parameters depending on (e.g.: nested function calls).
   */
  Evaluator getEvaluator(FunctionCall fc, Context context);
  /**
   * Return the <code>Evaluator</code> for the specified
   * <code>PredefinedFunction</code>.
   * <p>
   * Note that a <code>PredefinedFunction</code> is no
   * <code>Expression</code>. It's necessary to call this method
   * after {@link #getEvaluator(FunctionCall,Context)} has created
   * a new {@link FunctionCallEvaluator} which itself calls
   * {@link PredefinedFunction#getCallEvaluator(Type,Context)} -
   * a <code>PredefinedFunction</code> call preconditioned.
   * </p>
   * 
   * @return the <code>Evaluator</code> for the specified
   * <code>PredefinedFunction</code>.
   * @param pf the <code>PredefinedFunction</code> the <code>Evaluator</code>
   * is to be returned for.
   * @param paramEval the <code>Evaluator</code> for the
   * function's parameter.
   */
  Evaluator getEvaluator(PredefinedFunction pf, Evaluator paramEval);
  /** 
   * Return the <code>Class</code> this <code>Type</code> is representing.
   * 
   * @return the <code>Class</code> this <code>Type</code> is representing.
   */
  Class getRepresentationClass();
  /**
   * Create and return a new <code>Variable</code>.
   * 
   * @param name the name of the <code>Variable</code> to be created.
   * @param value the value of the <code>Variable</code> to be created.
   * @return a new <code>Variable</code>.
   */
  abstract public Variable createVariable(String name, Object value);
}
