/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2003-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.mathExpr.evaluator.complexEvaluator;

import de.jtem.mathExpr.PredefinedFunction;
import de.jtem.mathExpr.evaluator.AbstractType;
import de.jtem.mathExpr.evaluator.Constant;
import de.jtem.mathExpr.evaluator.Context;
import de.jtem.mathExpr.evaluator.Evaluator;
import de.jtem.mathExpr.evaluator.Type;
import de.jtem.mathExpr.evaluator.Variable;
import de.jtem.mathExpr.evaluator.complexVectorEvaluator.ComplexTimesComplexVectorEvaluator;
import de.jtem.mathExpr.evaluator.complexVectorEvaluator.ComplexTimesRealVectorEvaluator;
import de.jtem.mathExpr.evaluator.complexVectorEvaluator.ComplexVectorEvaluator;
import de.jtem.mathExpr.evaluator.complexVectorEvaluator.ComplexVectorType;
import de.jtem.mathExpr.evaluator.realEvaluator.Real;
import de.jtem.mathExpr.evaluator.realEvaluator.RealType;
import de.jtem.mathExpr.evaluator.realVectorEvaluator.RealVectorType;
import de.jtem.mathExpr.parser.BinaryOperation;
//for doc-links:
import de.jtem.mathExpr.parser.Expression;
import de.jtem.mathExpr.parser.FunctionCall;
import de.jtem.mathExpr.parser.List;
import de.jtem.mathExpr.parser.Symbol;
import de.jtem.mathExpr.parser.UnaryOperation;
import de.jtem.mfc.field.Complex;

/**
 * The evaluation {@link Type} when a certain part of an {@link Expression}
 * is a {@link Complex} when evaluated.
 * The following list shows, what part must be <code>Complex</code> so that
 * {@link ComplexType#TYPE} is used to get the appropriate Evaluator:<br>
 * <ul>
 * <li>{@link BinaryOperation}: the first operand</li>
 * <li>{@link UnaryOperation}: the operand (there is only one)</li>
 * <li>{@link FunctionCall}: nothing (see {@link AbstractType})
 * <li>{@link List}: at least one entry while the remaining are {@link
 * Real}</li>
 * <li>{@link Symbol}: the appropriate {@link Variable} object must not be of
 * <code>Type ComplexType</code>. If the <code>Symbol</code> is a constant
 * double value, a new {@link Constant} object of <code>Type ComplexType</code>
 * will be instantiated. If the <code>Symbol</code> is named "i", then the
 * <code>Constant</code> {@link Constant#I} will be returned.
 * In any other case, the appropriate <code>Variable</code> (or
 * even <code>Constant</code>, if "pi" or something else)
 * will be returned independent from its <code>Type</code>.</li>
 * </ul>
 *
 * @see Symbol#isConstant()
 * @see Constant#getConstant(Symbol,Type)
 * @see Variable#getReturnType()
 */
public class ComplexType extends AbstractType
{
	/**
	 * A <code>static</code> instance of this class. The {@link #ComplexType()
	 * Constructor} of this class is <code>protected</code>, so you have to use
	 * this <code>static</code> field.
	 */
  static public final ComplexType TYPE=new ComplexType();
	/**
	 * Creates a new <code>ComplexType</code> instance. This constructor is
	 * <code>protected</code> because you do not need to create several
	 * instances of this class. Use the <code>static</code> field {@link
	 * #TYPE} instead.
	 */
  protected ComplexType(){}
	/**
	 * Returns a {@link Constant} or {@link Variable} represented by the
	 * specified <code>Symbol</code>. If the <code>Symbol</code> is a constant
   * double value, a new {@link Constant} object of <code>Type
   * ComplexType</code> will be instantiated. The specified <code>Context</code>
	 * contains the <code>Variable</code> if the <code>Symbol</code> is not a
	 * constant double value.<br><br>
	 * If the <code>Symbol</code> is named "i", then the
 	 * <code>Constant</code> {@link Constant#I} will be returned.
	 * If the <code>Symbol</code> is a mathematical constant like "pi" or "e", the
	 * appropriate <code>Constant</code> object will be returned by
	 * {@link Constant#getConstant(Symbol,Type)}.
	 * 
	 * @see Symbol#isConstant()
	 * @param context the <code>Context</code>, the <code>Evaluator<code> for
	 * the <code>Symbol</code> is getting from (in cases of representing a variable
	 * and not a constant).
	 */
  public Evaluator getEvaluator(Symbol expr, Context context)
  {
    if(expr.isConstant())
      return new ComplexConstant(expr.toString(), 
          new Complex(expr.getConstantValue()), this);
    Constant c=Constant.getConstant(expr, this);
    if(c!=null)
      return c;
    c=Constant.getConstant(expr, RealType.TYPE);
    if(c!=null)
      return c;
    return context.getEvaluator(expr);
    
  }
	/**
	 * Returns the <code>Evaluator</code> for the specified
	 * <code>UnaryOperation</code> which operand is a {@link
	 * Complex} when evaluated. The returned <code>Evaluator</code> will
	 * evaluate to a <code>ComplexType</code> result again.
	 * 
	 * @see UnaryOperation#getEvaluator(Type,Context)
	 * @see PositiveSignComplexEvaluator
	 * @see NegativeSignComplexEvaluator
	 * @throws UnsupportedOperationException if the specified
	 * <code>UnaryOperation</code> has an unknown type (see
	 * {@link UnaryOperation#getType()}).
	 */
	public Evaluator getEvaluator(UnaryOperation expr, Evaluator e)
	{
		char uoType=(char)expr.getType();
		switch(uoType)
		{
			case '+': return new PositiveSignComplexEvaluator(e);
			case '-': return new NegativeSignComplexEvaluator(e);
			default: throw new UnsupportedOperationException((char)expr.getType()+
					": Unknown unary operation.");
		}
	}
	/**
	 * Returns the <code>Evaluator</code> for the specified
	 * <code>BinaryOperation</code> which first operand is a {@link
	 * Complex} when evaluated. The returned <code>Evaluator</code>
	 * must <b>not</b> evaluate to a <code>ComplexType</code> result.
	 *
	 * @see BinaryOperation#getEvaluator(Type,Context)
	 * @see ComplexPlusRealEvaluator
	 * @see ComplexMinusRealEvaluator
	 * @see ComplexTimesRealEvaluator
	 * @see ComplexDivideRealEvaluator
	 * @see ComplexPowerRealEvaluator
	 * @see ComplexPlusComplexEvaluator
	 * @see ComplexMinusComplexEvaluator
	 * @see ComplexTimesComplexEvaluator
	 * @see ComplexDivideComplexEvaluator
	 * @see ComplexPowerComplexEvaluator
	 * @see ComplexTimesRealVectorEvaluator
	 * @see ComplexTimesComplexVectorEvaluator
   * @throws UnsupportedOperationException if the specified
	 * <code>BinaryOperation</code> has an unknown type (see
	 * {@link BinaryOperation#getType()}) or a certain binary operation
	 * is not supported (e.g.: Complex / ComplexVector).
	 */
  public Evaluator getEvaluator(BinaryOperation expr,
      Evaluator e1, Evaluator e2)
  {
    Type secondOpType=e2.getReturnType();
    char boType=(char)expr.getType();//+, -, * ...
    if(secondOpType==RealType.TYPE)
    {
      switch(boType)
      {
        case '+': return new ComplexPlusRealEvaluator(e1,e2);
        case '-': return new ComplexMinusRealEvaluator(e1,e2);
        case '*': return new ComplexTimesRealEvaluator(e1,e2);
        case '/': return new ComplexDivideRealEvaluator(e1,e2);
        case '^': return new ComplexPowerRealEvaluator(e1,e2);
        default:
          throw new UnsupportedOperationException("Complex "+boType+
              " Real: Operation is not supported.");
      }
    }
    else if(secondOpType==ComplexType.TYPE)
    {
      switch(boType)
      {
        case '+': return new ComplexPlusComplexEvaluator(e1,e2);
        case '-': return new ComplexMinusComplexEvaluator(e1,e2);
        case '*': return new ComplexTimesComplexEvaluator(e1,e2);
        case '/': return new ComplexDivideComplexEvaluator(e1,e2);
        case '^': return new ComplexPowerComplexEvaluator(e1,e2);
        default:
          throw new UnsupportedOperationException("Complex "+boType+
              " Complex: Operation is not supported.");
      }
    }
    else if(secondOpType==RealVectorType.TYPE)
    {
      switch(boType)
      {
        case '*': return new ComplexTimesRealVectorEvaluator(e1,e2);
        default:
          throw new UnsupportedOperationException("Complex "+boType+
              " RealVector: Operation is not supported.");
      }
    }
    else if(secondOpType==ComplexVectorType.TYPE)
    {
      switch(boType)
      {
        case '*': return new ComplexTimesComplexVectorEvaluator(e1,e2);
        default:
          throw new UnsupportedOperationException("Complex "+boType+
              " ComplexVector: Operation is not supported.");
      }
    }
    //else if(secondOpType==RealMatrixType.TYPE)
    //{
    //  switch(boType)
    //  {
    //    default:
    //      throw new UnsupportedOperationException("Complex "+boType+
    //          " RealMatrix: Operation is not supported.");
    //  }
    //}
    //else if(secondOpType==ComplexMatrixType.TYPE)
    //{
    //  switch(boType)
    //  {
    //    default:
    //      throw new UnsupportedOperationException("Complex "+boType+
    //          " ComplexMatrix: Operation is not supported.");
    //  }
    //}
    throw new UnsupportedOperationException(boType+
        ": Operation is not supported.");
  }
	/**
	 * Returns the <code>Evaluator</code> for the specified <code>List</code>
	 * which all entries are either {@link Complex} or {@link Real} when
	 * evaluated (at least one entry must be <code>Complex</code>).
	 * The <code>List</code> represents a {@link de.jtem.blas.ComplexVector} in any case
	 * and the returned <code>Evaluator</code> evaluates to a
	 * <code>ComplexVector</code> result.
	 * 
	 * @see List#getEvaluator(Type,Context)
	 * @return a new {@link ComplexVectorEvaluator} instance
	 */
	public Evaluator getEvaluator(List expr, Evaluator[] e)
	{
		return new ComplexVectorEvaluator(expr, e); 
	}
	/**
	 * Returns the <code>Evaluator</code> for the specified
	 * <code>PredefinedFunction</code> which parameter is a {@link
	 * Complex} when evaluated.
	 *
	 * @see PredefinedFunctionComplexEvaluator
   * @throws UnsupportedOperationException if the specified
   * <code>PredefinedFunction</code> is unknown.
	 */
  public Evaluator getEvaluator(PredefinedFunction pf, Evaluator paramEval)
  {
    int i;
    for(i=0; i<PredefinedFunction.FUNCTIONS.length; i++)
      if(pf==PredefinedFunction.FUNCTIONS[i])
        break;
    switch(i)
    {
      case 0://sin
        return new PredefinedFunctionComplexEvaluator.SinEvaluator(paramEval);
      case 1://cos
        return new PredefinedFunctionComplexEvaluator.CosEvaluator(paramEval);
      case 2://tan
      	return new PredefinedFunctionComplexEvaluator.TanEvaluator(paramEval);
      case 3://exp
        return new PredefinedFunctionComplexEvaluator.ExpEvaluator(paramEval);
      case 4://log
        return new PredefinedFunctionComplexEvaluator.LogEvaluator(paramEval);
      case 5://sqrt
        return new PredefinedFunctionComplexEvaluator.SqrtEvaluator(paramEval);
      case 6://atan
      	throw new UnsupportedOperationException( "complex atan not supported" );
      default:
			  throw new UnsupportedOperationException(pf.getName()+
					  ": Unknown predefined function.");
    }
  }
	/**
	 * Returns {@link Complex}<code>.class</code>.
	 * 
	 * @return <code>Complex.class</code>.
	 */  
	public Class getRepresentationClass()
  {
    return Complex.class;
  }
	/**
	 * Returns <code>true</code> if <code>o</code> is {@link #TYPE},
	 * <code>false</code> otherwise.
	 * 
	 * @param o the <code>Object</code> to compare this <code>ComplexType</code>
	 * against.
	 * @return <code>true</code> if <code>o</code> is {@link #TYPE},
	 * <code>false</code> otherwise.
	 */
	public boolean equals(Object o)
	{
		return this==o;
	}
  /**
   * Creates and returns a new {@link ComplexVariable}.
   * 
   * @see Type#createVariable(String,Object)
   * @return a new <code>ComplexVariable</code>.
   */
  public Variable createVariable(String name, Object value)
  {
    return new ComplexVariable(name, (Complex)value, this);
  }
}
