/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2003-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.mathExpr.evaluator.realEvaluator;

//for doc-links:
import de.jtem.mathExpr.PredefinedFunction;
import de.jtem.mathExpr.evaluator.Evaluator;

/**
 * An abstraction of all <code>Evaluators</code> evaluating {@link
 * PredefinedFunction PredefinedFunctions} which <b>function call parameter
 * is a {@link Real}</b> when evaluated - in other words:
 * This class provides function calls of predefined functions on real values.
 * This class provides <code>static</code> nested classes inheriting this class
 * - one <code>static</code> class provides the evaluation of the appropriate
 * <code>PredefinedFunction</code>.
 * All you have to do is to create an instance of such <code>static</code> class
 * and give the constructor the parameter <code>Evaluator</code>. The method
 * {@link #evaluate()} evaluates that parameter and executes the predefined
 * function call on the evaluation result.
 * 
 * @author marcel
 */
public abstract class PredefinedFunctionRealEvaluator
		extends AbstractRealEvaluator
{
  /**
   * The <code>Evaluator</code> of the function call parameter.
   */
  Evaluator paramEval;
  /**
   * Creates a new <code>PredefinedFunctionRealEvaluator</code>
   * executing the certain function call on the evaluation result of
   * <code>e</code>.
   * 
   * @param e the <code>Evaluator</code> of the function call parameter.
   */
  public PredefinedFunctionRealEvaluator(Evaluator e)
  {
    paramEval=e;
  }
	/**
	 * Execute the certain function call on the evaluation result of the given
	 * parameter <code>Evaluator</code> and return the result as a {@link Real}
	 * <code>Object</code>.
	 * 
	 * @return the result of the function call on the parameter as a
	 * {@link Real} <code>Object</code>. 
	 * @see #getReturnType()
	 */
  abstract public Object evaluate();
  /**
   * The <code>PredefinedFunctionRealEvaluator</code> executing
   * sine on a given real parameter.
   */
  static public class SinEvaluator extends PredefinedFunctionRealEvaluator
  {
  	/**
  	 * Creates a new <code>SinEvaluator</code> with the parameter
  	 * <code>Evaluator e</code>.
  	 * 
  	 * @param e the <code>Evaluator</code> from its real evaluation result
  	 * sine is to be calculated.
  	 */
    public SinEvaluator(Evaluator e)
    {
      super(e);
    }
    /**
     * Executes sine on the real evaluation result of the given parameter
     * <code>Evaluator</code> and returns the result as a {@link Real}
     * <code>Object</code>.
     * 
     * @return the sine of the parameter as a {@link Real} <code>Object</code>. 
     * @see #getReturnType()
     */
    public Object evaluate()
    {
      realValue.assignSin((Real)paramEval.evaluate());
      return realValue;
    }
  }
	/**
	 * The <code>PredefinedFunctionRealEvaluator</code> executing
	 * cosine on a given real parameter.
	 */
  static public class CosEvaluator extends PredefinedFunctionRealEvaluator
  {
		/**
		 * Creates a new <code>CosEvaluator</code> with the parameter
		 * <code>Evaluator e</code>.
		 * 
		 * @param e the <code>Evaluator</code> from its real evaluation result
		 * cosine is to be calculated.
		 */
    public CosEvaluator(Evaluator e)
    {
      super(e);
    }
		/**
		 * Executes cosine on the real evaluation result of the given parameter
		 * <code>Evaluator</code> and returns the result as a {@link Real}
		 * <code>Object</code>.
		 * 
		 * @return the cosine of the parameter as a {@link Real} <code>Object</code>. 
		 * @see #getReturnType()
		 */
    public Object evaluate()
    {
      realValue.assignCos((Real)paramEval.evaluate());
      return realValue;
    }
  }
  /**
   * The <code>PredefinedFunctionRealEvaluator</code> executing
   * tangent on a given real parameter.
   */
  static public class TanEvaluator extends PredefinedFunctionRealEvaluator
  {
  	/**
  	 * Creates a new <code>TanEvaluator</code> with the parameter
  	 * <code>Evaluator e</code>.
  	 * 
  	 * @param e the <code>Evaluator</code> from its real evaluation result
  	 * tagent is to be calculated.
  	 */
  	public TanEvaluator(Evaluator e)
	{
  		super(e);
  	}
  	/**
  	 * Executes tangent on the real evaluation result of the given parameter
  	 * <code>Evaluator</code> and returns the result as a {@link Real}
  	 * <code>Object</code>.
  	 * 
  	 * @return the tangent of the parameter as a {@link Real} <code>Object</code>. 
  	 * @see #getReturnType()
  	 */
  	public Object evaluate()
	{
  		realValue.assignTan((Real)paramEval.evaluate());
  		return realValue;
  	}
  }
	/**
	 * The <code>PredefinedFunctionRealEvaluator</code> executing
	 * exponential function on a given real parameter.
	 */
  static public class ExpEvaluator extends PredefinedFunctionRealEvaluator
  {
		/**
		 * Creates a new <code>ExpEvaluator</code> with the parameter
		 * <code>Evaluator e</code>.
		 * 
		 * @param e the <code>Evaluator</code> on its real evaluation result
		 * exponential function is to be executed.
		 */
    public ExpEvaluator(Evaluator e)
    {
      super(e);
    }
		/**
		 * Executes exponential function on the real evaluation result of the given
		 * parameter <code>Evaluator</code> and returns the result as a {@link Real}
		 * <code>Object</code>.
		 * 
		 * @return the exponential function result of the parameter as a
		 * {@link Real} <code>Object</code>. 
		 * @see #getReturnType()
		 */
    public Object evaluate()
    {
      realValue.assignExp((Real)paramEval.evaluate());
      return realValue;
    }
  }
	/**
	 * The <code>PredefinedFunctionRealEvaluator</code> executing
	 * natural logarithm on a given real parameter.
	 */
  static public class LogEvaluator extends PredefinedFunctionRealEvaluator
  {
		/**
		 * Creates a new <code>LogEvaluator</code> with the parameter
		 * <code>Evaluator e</code>.
		 * 
		 * @param e the <code>Evaluator</code> from its real evaluation result
		 * the natural logarithm is to be calculated.
		 */
    public LogEvaluator(Evaluator e)
    {
      super(e);
    }
		/**
		 * Executes natural logarithm on the real evaluation result of the given
		 * parameter <code>Evaluator</code> and returns the result as a {@link Real}
		 * <code>Object</code>.
		 * 
		 * @return the natural logarithm of the parameter as a {@link Real}
		 * <code>Object</code>. 
		 * @see #getReturnType()
		 */
    public Object evaluate()
    {
      realValue.assignLog((Real)paramEval.evaluate());
      return realValue;
    }
  }
	/**
	 * The <code>PredefinedFunctionRealEvaluator</code> executing
	 * square root on a given real parameter.
	 */
  static public class SqrtEvaluator extends PredefinedFunctionRealEvaluator
  {
		/**
		 * Creates a new <code>SqrtEvaluator</code> with the parameter
		 * <code>Evaluator e</code>.
		 * 
		 * @param e the <code>Evaluator</code> from its real evaluation result
		 * square root is to be calculated.
		 */
    public SqrtEvaluator(Evaluator e)
    {
      super(e);
    }
		/**
		 * Executes square root on the real evaluation result of the given parameter
		 * <code>Evaluator</code> and returns the result as a {@link Real}
		 * <code>Object</code>.
		 * 
		 * @return the square root of the parameter as a {@link Real}
		 * <code>Object</code>. 
		 * @see #getReturnType()
		 */
    public Object evaluate()
    {
      realValue.assignSqrt((Real)paramEval.evaluate());
      return realValue;
    }
  }
  /**
   * The <code>PredefinedFunctionRealEvaluator</code> executing
   * atan on a given real parameter.
   */
  static public class ATanEvaluator extends PredefinedFunctionRealEvaluator
  {
  	/**
  	 * Creates a new <code>ATanEvaluator</code> with the parameter
  	 * <code>Evaluator e</code>.
  	 * 
  	 * @param e the <code>Evaluator</code> from its real evaluation result
  	 * atan is to be calculated.
  	 */
  	public ATanEvaluator(Evaluator e)
	{
  		super(e);
  	}
  	/**
  	 * Executes atan on the real evaluation result of the given parameter
  	 * <code>Evaluator</code> and returns the result as a {@link Real}
  	 * <code>Object</code>.
  	 * 
  	 * @return the atan of the parameter as a {@link Real}
  	 * <code>Object</code>. 
  	 * @see #getReturnType()
  	 */
  	public Object evaluate()
	{
  		realValue.assignATan((Real)paramEval.evaluate());
  		return realValue;
  	}
  }
}
