/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2003-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.mathExpr.evaluator.realEvaluator;

//for doc-links:
import de.jtem.mathExpr.evaluator.Evaluator;
import de.jtem.mathExpr.parser.Expression;

/**
 * A wrapper class for any <code>double</code> value. In opposite to the
 * wrapper class {@link Double}, in this class the value can be changed any
 * time. This class has methods like {@link de.jtem.mfc.field.Complex} to provide
 * mathematical operations on the value.
 * <br><br>
 * This class is necessary for certain {@link Evaluator Evaluators} to provide
 * working on <code>Objects</code> if the {@link Expression}
 * contains double values.
 * 
 * @see RealType
 */
public class Real
{
  private double value;
  /**
   * Creates a new <code>Real</code> object with value==0.0.
   */
  public Real()
  {
    this(0.);
  }
  public Real(Real initValue)
  {
    this(initValue.getValue());
  }
  public Real(double initValue)
  {
    value=initValue;
  }
  //--------------------------------
  public void assign(Real newValue)
  {
    if(value!=newValue.getValue())
      value=newValue.getValue();
  }
  public void assign(double newValue)
  {
    if(value!=newValue)
      value=newValue;
  }
  public double getValue()
  {
    return value;
  }
  public boolean equals(Object obj)
  {
    if(!(obj instanceof Real))
      return false;
    Real real=(Real)obj;
    return value==real.getValue();
  }
  public String toString()
  {
    return String.valueOf(value);
  }
  //--------------------------------
  public void assignNeg()
  {
    value=-value;
  }
  public void assignNeg(Real real)
  {
    value=-real.getValue();
  }
  public void assignPlus(Real addend)
  {
    value+=addend.getValue();
  }
  public void assignPlus(Real addend1, Real addend2)
  {
    value=addend1.getValue()+addend2.getValue();
  }
  public void assignPlus(double addend)
  {
    value+=addend;
  }
  public void assignMinus(Real subtrahend)
  {
    value-=subtrahend.getValue();
  }
  public void assignMinus(Real minuend, Real subtrahend)
  {
    value=minuend.getValue()-subtrahend.getValue();
  }
  public void assignMinus(double subtrahend)
  {
    value-=subtrahend;
  }
  public void assignTimes(Real multiplier)
  {
    value*=multiplier.getValue();
  }
  public void assignTimes(Real multiplier1, Real multiplier2)
  {
    value=multiplier1.getValue()*multiplier2.getValue();
  }
  public void assignTimes(double multiplier)
  {
    value*=multiplier;
  }
  public void assignDivide(Real denominator)
  {
    value/=denominator.getValue();
  }
  public void assignDivide(Real nominator, Real denominator)
  {
    value=nominator.getValue()/denominator.getValue();
  }
  public void assignDivide(double denominator)
  {
    value/=denominator;
  }
  public void assignPow(Real exponent)
  {
    value=Math.pow(value, exponent.getValue());
  }
  public void assignPow(Real basis, Real exponent)
  {
    value=Math.pow(basis.getValue(), exponent.getValue());
  }
  public void assignPow(double exponent)
  {
    value=Math.pow(value, exponent);
  }
  public void assignSqrt()
  {
    value=Math.sqrt(value);
  }
  public void assignSqrt(Real real)
  {
    value=Math.sqrt(real.getValue());
  }
  public void assignATan(Real real)
  {
  	value=Math.atan(real.getValue());
  }
  public void assignSin()
  {
    value=Math.sin(value);
  }
  public void assignSin(Real real)
  {
    value=Math.sin(real.getValue());
  }
  public void assignCos()
  {
    value=Math.cos(value);
  }
  public void assignCos(Real real)
  {
    value=Math.cos(real.getValue());
  }
  public void assignTan(Real real)
  {
  	value=Math.tan(real.getValue());
  }
  public void assignLog()
  {
    value=Math.log(value);
  }
  public void assignLog(Real real)
  {
    value=Math.log(real.getValue());
  }
  public void assignExp()
  {
    value=Math.exp(value);
  }
  public void assignExp(Real real)
  {
    value=Math.exp(real.getValue());
  }
  //--------------------------------
  public Real plus(Real addend)
  {
    return new Real(value+addend.getValue());
  }
  public Real plus(double addend)
  {
    return new Real(value+addend);
  }
  public Real minus(Real subtrahend)
  {
    return new Real(value-subtrahend.getValue());
  }
  public Real minus(double subtrahend)
  {
    return new Real(value-subtrahend);
  }
  public Real times(Real multiplier)
  {
    return new Real(value*multiplier.getValue());
  }
  public Real times(double multiplier)
  {
    return new Real(value*multiplier);
  }
  public Real divide(Real denominator)
  {
    return new Real(value/denominator.getValue());
  }
  public Real divide(double denominator)
  {
    return new Real(value/denominator);
  }
  public Real pow(Real exponent)
  {
    return new Real(Math.pow(value,exponent.getValue()));
  }
  public Real pow(double exponent)
  {
    return new Real(Math.pow(value,exponent));
  }
  public Real sqrt()
  {
    return new Real(Math.sqrt(value));
  }
  public Real sin()
  {
    return new Real(Math.sin(value));
  }
  public Real cos()
  {
    return new Real(Math.cos(value));
  }
  public Real log()
  {
    return new Real(Math.log(value));
  }
  public Real exp()
  {
    return new Real(Math.exp(value));
  }
}
