/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2003-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.mathExpr.evaluator.realVectorEvaluator;

import de.jtem.blas.RealVector;
import de.jtem.mathExpr.PredefinedFunction;
import de.jtem.mathExpr.evaluator.AbstractType;
import de.jtem.mathExpr.evaluator.Context;
import de.jtem.mathExpr.evaluator.Evaluator;
import de.jtem.mathExpr.evaluator.Type;
import de.jtem.mathExpr.evaluator.Variable;
import de.jtem.mathExpr.evaluator.complexEvaluator.ComplexType;
import de.jtem.mathExpr.evaluator.realEvaluator.RealType;
import de.jtem.mathExpr.parser.BinaryOperation;
//for doc-links:
import de.jtem.mathExpr.parser.Expression;
import de.jtem.mathExpr.parser.FunctionCall;
import de.jtem.mathExpr.parser.List;
import de.jtem.mathExpr.parser.Symbol;
import de.jtem.mathExpr.parser.UnaryOperation;

/**
 * The evaluation {@link Type} when a certain part of an {@link Expression}
 * is a {@link RealVector} when evaluated.
 * The following list shows, what part must be a <code>RealVector</code> so that
 * {@link RealVectorType#TYPE} is used to get the appropriate Evaluator:<br>
 * <ul>
 * <li>{@link BinaryOperation}: the first operand</li>
 * <li>{@link UnaryOperation}: the operand (there is only one)</li>
 * <li>{@link FunctionCall}: nothing (see {@link AbstractType})
 * <li>{@link List}: <i>not implemented yet</i></li>
 * <li>{@link Symbol}: <i>not implemented yet</i></li>
 * </ul>
 */
public class RealVectorType extends AbstractType
{
	/**
	 * A <code>static</code> instance of this class. The {@link #RealVectorType()
	 * Constructor} of this class is <code>protected</code>, so you have to use
	 * this <code>static</code> field.
	 */
  static public final RealVectorType TYPE=new RealVectorType();
	/**
	 * Creates a new <code>RealVectorType</code> instance. This constructor is
	 * <code>protected</code> because you do not need to create several
	 * instances of this class. Use the <code>static</code> field {@link
	 * #TYPE} instead.
	 */
  protected RealVectorType(){}
  /**
   * Throws an <code>UnsupportedOperationException</code>
   * because this method is not implemented yet. Use {@link RealType} or
   * {@link ComplexType} as your default evaluator {@link Type} to avoid
   * calling this method.
   *
   * @throws UnsupportedOperationException whenever this method is called.
   * @return nothing, because method is not implemented.
   */
  public Evaluator getEvaluator(Symbol expr, Context context)
  {
  	//TODO: implementing this method. I'm not sure how to do so -
  	//evaluate Symbols with a constant double value to a RealVector? 
    throw new UnsupportedOperationException("Not implemented yet. "+
				"Use RealType or ComplexType as your default Type.");
  }
	/**
	 * Returns the <code>Evaluator</code> for the specified
	 * <code>UnaryOperation</code> which operand is a {@link RealVector}
	 * when evaluated. The returned <code>Evaluator</code> will
	 * evaluate to a <code>RealVectorType</code> result again.
	 * 
	 * @see UnaryOperation#getEvaluator(Type,Context)
	 * @see PositiveSignRealVectorEvaluator
	 * @see NegativeSignRealVectorEvaluator
	 * @throws UnsupportedOperationException if the specified
	 * <code>UnaryOperation</code> has an unknown type (see
	 * {@link UnaryOperation#getType()}).
	 */
	public Evaluator getEvaluator(UnaryOperation expr, final Evaluator e)
	{
		char uoType=(char)expr.getType();
		switch(uoType)
		{
			case '+': return new PositiveSignRealVectorEvaluator(e);
			case '-': return new NegativeSignRealVectorEvaluator(e);
			default: throw new UnsupportedOperationException((char)expr.getType()+
					": Unknown unary operation.");
		}
	}
	/**
	 * Returns the <code>Evaluator</code> for the specified
	 * <code>BinaryOperation</code> which first operand is a {@link RealVector}
	 * when evaluated. The returned <code>Evaluator</code> must <b>not</b>
	 * evaluate to a <code>RealVectorType</code> result.
	 *
	 * @see BinaryOperation#getEvaluator(Type,Context)
	 * @see RealVectorTimesRealEvaluator
	 * @see RealVectorDivideRealEvaluator
	 * @see RealVectorPlusRealVectorEvaluator
	 * @see RealVectorMinusRealVectorEvaluator
   * @throws UnsupportedOperationException if the specified
	 * <code>BinaryOperation</code> has an unknown type (see
	 * {@link BinaryOperation#getType()}) or a certain binary operation
	 * is not supported (e.g.: RealVector + Real).
   */
  public Evaluator getEvaluator(BinaryOperation expr,
      Evaluator e1, Evaluator e2)
  {
    Type secondOpType=e2.getReturnType();
    char boType=(char)expr.getType();//+, -, * ...
    if(secondOpType==RealType.TYPE)
    {
      switch(boType)
      {
        case '*': return new RealVectorTimesRealEvaluator(e1,e2);
        case '/': return new RealVectorDivideRealEvaluator(e1,e2);
        default:
          throw new UnsupportedOperationException("RealVector "+boType+
              " Real: Operation is not supported.");
      }
    }
    else if(secondOpType==RealVectorType.TYPE)
    {
      switch(boType)
      {
        case '+': return new RealVectorPlusRealVectorEvaluator(e1,e2);
        case '-': return new RealVectorMinusRealVectorEvaluator(e1,e2);
        default:
          throw new UnsupportedOperationException("RealVector "+boType+
              " RealVector: Operation is not supported.");
      }
    }
    throw new UnsupportedOperationException(boType+
        ": Operation is not supported.");
  }
	/**
	 * Throws an <code>UnsupportedOperationException</code>, because this method
	 * is not implemented yet.
	 * 
   * @throws UnsupportedOperationException whenever this method is called.
   * @return nothing, because method is not implemented.
	 */
	public Evaluator getEvaluator(List expr, Evaluator[] e)
	{
		throw new UnsupportedOperationException("Not implemented yet.");
	}
  /**
 	 * Throws an <code>UnsupportedOperationException</code>, because this method
	 * is not implemented yet.
   * 
   * @throws UnsupportedOperationException whenever this method is called.
   * @return nothing, because method is not implemented.
   */
  public Evaluator getEvaluator(PredefinedFunction pf, Evaluator paramEval)
  {
    throw new UnsupportedOperationException("Not implemented yet.");
  }
	/**
	 * Returns {@link RealVector}<code>.class</code>.
	 * 
	 * @return <code>RealVector.class</code>.
	 */
  public Class getRepresentationClass()
  {
    return RealVector.class;
  }
	/**
	 * Returns <code>true</code> if <code>o</code> is {@link #TYPE},
	 * <code>false</code> otherwise.
	 * 
	 * @param o the <code>Object</code> to compare this
	 * <code>RealVectorType</code> against.
	 * @return <code>true</code> if <code>o</code> is {@link #TYPE},
	 * <code>false</code> otherwise.
	 */
	public boolean equals(Object o)
	{
		return this==o;
	}
  /**
   * Not implemented yet.
   */
  public Variable createVariable(String name, Object value)
  {
    throw new UnsupportedOperationException("RealVectorVariables can't be" +
      "created yet");
  }
}
