/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2003-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.mathExpr.parser;

import java.util.Collection;

import de.jtem.mathExpr.evaluator.Context;
import de.jtem.mathExpr.evaluator.Evaluator;
import de.jtem.mathExpr.evaluator.Type;
//for doc-links:
import de.jtem.mathExpr.evaluator.Variable;
import de.jtem.mathExpr.evaluator.realEvaluator.Real;
import de.jtem.mathExpr.parser.javacc.Parser;

/**
 * A <code>BinaryOperation</code> is an <code>Expression</code> which
 * represents a mathematical binary operation (addition, subtraction,
 * multiplication, division or power operation) of two other
 * <code>Expressions</code> (the operands).
 */
public abstract class BinaryOperation extends Expression
{
  /**
   * This <code>BinaryOperation's</code> first operand.
   */
  protected Expression operand1;
  /**
   * This <code>BinaryOperation's</code> second operand.
   */
  protected Expression operand2;
  /**
   * The operation type of this <code>BinaryOperation</code>.
   * 
   * @see #getType()
   */
  private int type;//TODO: add appropriate final int values in this class, marcel!
  /**
   * The evaluation priority of this BinaryOperation.
   */
  private int priority;
  /**
   * Creates a new <code>BinaryOperation</code> with the operation type
   * <code>type</code> and the evaluation priority <code>pri</code>.
   * The operands of this <code>BinaryOperation</code> will be
   * <code>null</code> until you call the method {@link
   * #setOperands(Expression,Expression)}. Use the other constructor, if
   * the operands are already known.
   * 
   * @param type the operation type this <code>BinaryOperation</code> will
   * represent.
   * @param pri the evaluation priority this <code>BinaryOperation</code> will
   * have.
   */
  public BinaryOperation(int type, int pri)
  {
    this.type=type;
    priority=pri;
  }
  /**
   * Creates a new <code>BinaryOperation</code> with the operation type
   * <code>type</code> , the evaluation priority <code>pri</code> and both
   * operands <code>op1</code> and <code>op2</code>. Use the other
   * constructor if at least one of the operands is not known.
   * 
   * @param type the operation type this <code>BinaryOperation</code> will
   * represent.
   * @param pri the evaluation priority this <code>BinaryOperation</code> will
   * have.
   * @param op1 the first operand for this <code>BinaryOperation</code>.
   * @param op2 the second operand for this <code>BinaryOperation</code>.
   */
  public BinaryOperation(int type, int pri, Expression op1, Expression op2)
  {
    this(type, pri);
    setOperands(op1, op2);
  }
  /**
   * Sets the operands of this <code>BinaryOperation</code> to the specified
   * <code>Expressions</code>.
   * 
   * @param op1 the new first operand for this <code>BinaryOperation</code>.
   * @param op2 the new second operand for this <code>BinaryOperation</code>.
   */
  public void setOperands(Expression op1, Expression op2)
  {
    operand1=op1;//TODO: maybe they are already equal, marcel!
    operand2=op2;
  }
  /**
   * Returns the first operand of this <code>BinaryOperation</code> or
   * <code>null</code> if there is no operand defined yet. Define the
   * operands with the appropriate constructor or with the method
   * {@link #setOperands(Expression, Expression)}.
   * 
   * @return the first operand of this <code>BinaryOperation</code>.
   */
  public Expression getOperand1()
  {
    return operand1;
  }
  /**
   * Returns the second operand of this <code>BinaryOperation</code> or
   * <code>null</code> if there is no operand defined yet. Define the
   * operands with the appropriate constructor or with the method
   * {@link #setOperands(Expression, Expression)}.
   * 
   * @return the second operand of this <code>BinaryOperation</code>.
   */
  public Expression getOperand2()
  {
    return operand2;
  }
  /**
   * Returns the operation type of this <code>BinaryOperation</code>.
   * It's the ASCII-code of the following characters representing the
   * appropriate operation:<br><br>
   * &nbsp;&nbsp;&nbsp;&nbsp;'+': addition<br>
   * &nbsp;&nbsp;&nbsp;&nbsp;'-': subtraction<br>
   * &nbsp;&nbsp;&nbsp;&nbsp;'*': multiplication<br>
   * &nbsp;&nbsp;&nbsp;&nbsp;'/': division<br>
   * &nbsp;&nbsp;&nbsp;&nbsp;'^': power
   * 
   * @return the operation type of this <code>BinaryOperation</code>.
   */
  public int getType()
  {
    return type;
  }
  /**
   * Returns the evaluation priority of this <code>BinaryOperation</code>.
   * 
   * @return the evaluation priority of this <code>BinaryOperation</code>.
   */
  public int getPriority()
  {
    return priority;
  }
  /**
   * Returns <code>true</code> if both operands of this
   * <code>BinaryOperation</code> are not null <b>and</b> also constant,
   * <code>false</code> if not.
   * 
   * @return <code>true</code> if both operands of this
   * <code>BinaryOperation</code> are not null <b>and</b> also constant,
   * <code>false</code> if not.
   */
  public boolean isConstant()
  {
    return (operand1!=null && operand1.isConstant())
        && (operand2!=null && operand2.isConstant());
  }
  /**
   * Returns <code>true</code> if both operands of this
   * <code>BinaryOperation</code> are not null <b>and</b> also complete,
   * <code>false</code> otherwise.
   * 
   * @return <code>true</code> if both operands of this
   * <code>BinaryOperation</code> are not null <b>and</b> also complete,
   * <code>false</code> otherwise.
   */
  public boolean isComplete()
  {
    return (operand1!=null && operand1.isComplete())
        && (operand2!=null && operand2.isComplete());
  }
  /**
   * Adds the names of the {@link Variable Variables} which are essential for
   * evaluation into the specified <code>Collection</code>. Therefor it gets the
   * essential <code>Variables</code> of both operands of this
   * <code>BinaryOperation</code>.
   * 
   * @param c the <code>Collection</code> the <code>Variable</code> names
   * will be added to.
   */
  public void getEssentialVariables(Collection c)
  {
    operand1.getEssentialVariables(c);
    operand2.getEssentialVariables(c);
  }
  /**
   * Returns the <code>Evaluator</code> for this <code>BinaryOperation</code>.
   * <code>Type defaultType</code> specifies, how to deal with
   * <code>Symbols</code> in the expression-tree-hirarchy (evaluate any
   * <code>Symbol</code> as a {@link de.jtem.mfc.field.Complex} or a {@link Real} or
   * whatever?). It has nothing to do with {@link #getType()}.
   * 
   * @return the <code>Evaluator</code> for this <code>BinaryOperation</code>.
   * @param defaultType the <code>Type</code> any <code>Symbol</code> will be
   * evaluated to.
   * @param context the <code>Context</code>, the <code>Evaluators<code> for
   * the <code>Symbols</code> are getting from.
   */
  public Evaluator getEvaluator(Type defaultType, Context context)
  {
    Evaluator e1=operand1.getEvaluator(defaultType, context);
    Evaluator e2=operand2.getEvaluator(defaultType, context);
    return e1.getReturnType().getEvaluator(this, e1, e2);
  }
  /**
   * Replaces all Symbols equal to s in this BinaryOperation
   * with the specified Expression. Therefor it replaces
   * all Symbols equal to s in this' operands.
   * 
   * @see Symbol#replaceSymbol(Symbol,Expression)
   */
  public Expression replaceSymbol(Symbol s, Expression expr)
  {
    BinaryOperation newBO=null;
    try {
      newBO=(BinaryOperation)(this.clone()); }
    catch(CloneNotSupportedException ex){}
    Expression newOp1, newOp2;
    newOp1=operand1.replaceSymbol(s, expr);
    newOp2=operand2.replaceSymbol(s, expr);
    newBO.setOperands(newOp1, newOp2);
    return newBO;
  }

  public Expression dissolveFunctionCalls()
  {
    BinaryOperation newBO=null;
    try {
      newBO=(BinaryOperation)(this.clone()); }
    catch(CloneNotSupportedException ex){}
    Expression newOp1=operand1.dissolveFunctionCalls();
    Expression newOp2=operand2.dissolveFunctionCalls();
    newBO.setOperands(newOp1, newOp2);
    return newBO;
  }
  /**
   * Returns <code>true</code> if the specified <code>Object</code> is a
   * <code>BinaryOperation</code> of the same operation type and with equal
   * operands, <code>false</code> otherwise.
   * 
   * @param o the object to compare this <code>BinaryOperation</code> against.
   * @return <code>true</code> if the specified <code>Object</code> is a
   * <code>BinaryOperation</code> of the same operation type and with equal
   * operands, <code>false</code> otherwise.
   */
  public boolean equals(Object o)
  {
    if(o==this) return true;
    if(!(o instanceof BinaryOperation))
      return false;
    BinaryOperation bo=(BinaryOperation)o;
    if(type!=bo.getType())
      return false;
    if(!(operand1.equals(bo.getOperand1()) &&
        operand2.equals(bo.getOperand2())))
      return false;
    return true;
  }
  /**
   * Returns a <code>String</code> representing this
   * <code>BinaryOperation</code>. It's now possible to parse the
   *  returned <code>String</code> by a {@link Parser} again.
   * 
   * @return a <code>String</code> representing this
   * <code>BinaryOperation</code>.
   */
  public String toString()
  {
    char op=(char)type;
    return "("+   (operand1!=null? operand1.toString(): "?")
      +' '+op+' '+(operand2!=null? operand2.toString(): "?")
      +")";
  }
}
