/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2003-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.mathExpr.parser;

import de.jtem.mathExpr.DefinitionModel;
//for doc-links:
import de.jtem.mathExpr.ExpressionConfiguration;
import de.jtem.mathExpr.Function;
import de.jtem.mathExpr.PredefinedFunction;
import de.jtem.mathExpr.UnknownDefinitionException;
import de.jtem.mathExpr.UserDefinedFunction;
import de.jtem.mathExpr.evaluator.Variable;
import de.jtem.mathExpr.parser.javacc.Parser;

/**
 * Provides methods to create most possible kinds of {@link Expression
 * Expressions}. For example:<br>
 * <ul>
 * <li>{@link BinaryOperation BinaryOperations} of the several types
 * '+', '-', '*', '/', '^' (power), e.g.: a+b, a-b, a*b, a/b, a^b</li>
 * <li>{@link UnaryOperation UnaryOperations} of the several types
 * '+', '-', e.g.: +a, -a</li>
 * <li>{@link FunctionCall FunctionCalls} of {@link PredefinedFunction
 * Predefined-} and {@link UserDefinedFunction UserDefinedFunctions}, e.g.:
 * f(a), sin(a)</li>
 * <li>{@link Symbol Symbols}: variable identifiers or constant values, e.g.: a, 1.</li>
 * </ul>
 * This class is particularly used by a {@link Parser} to create all needed
 * <code>Expressions</code> to build an expression-tree-hirarchy.
 * 
 * @see ExpressionConfiguration
 */
public class ExpressionFactory
{
  private DefinitionModel defModel;
  /**
   * Creates a new <code>ExpressionFactory</code> instance.
   * The specified <code>DefinitionModel</code> is important for this
   * <code>ExpressionFactory</code> to know
   * about all defined {@link Function Functions}. That provides the ability to
   * decide if a certain parsed identifier is a {@link Symbol} in an implicit
   * multiplication or a {@link FunctionCall}, e.g.:<br>
   * "f(2)" will be only parsed as a <code>FunctionCall</code>, if 'f' is
   * defined as a <code>Function</code> in the <code>DefintionModel</code>.
   * Otherwise it will be parsed to a <code>BinaryOperation</code> 'f*2'.
   * 
   * @see ExpressionConfiguration#defineFunction(String,String[],String)
   * @see DefinitionModel#containsFunction(String)
   * @see #createImplicitOperation(Expression)
   * @param defModel the <code>DefinitionModel</code> holding all
   * <code>Functions</code> comparing a certain parsed identifier against
   * to decide between a <code>FunctionCall</code> or a
   * <code>BinaryOperation</code>.
   */
  public ExpressionFactory(DefinitionModel defModel)
  {
    this.defModel=defModel;
  }
  /**
   * Creates and returns a new {@link BinaryOperation} with the type
   * <code>i</code> and a default priority.
   * 
   * @throws UnsupportedOperationException if <code>i</code> is no ASCII-Code
   * for '+', '-', '*' ,'/' or '^'.
   * @see BinaryOperation#getType()
   * @see Expression#getPriority()
   * @param i the ASCI-Code for the several <code>BinaryOperation</code> types
   * ('+', '-', '*', '/' or '^').
   * @return a new {@link BinaryOperation} with the type
   * <code>i</code> and a default priority.
   */
  public Expression createBinaryOperation(int i)
  {
    int pri;
    switch(i)
    {
    	//TODO: static public final variables for '+', '-' ... (marcel)
      case '+': return new Addition(); //pri=2; break;
      case '*': return new Multiplication(); //pri=4; break;
      case '-': return new Subtraction(); //pri=2; break;
      case '/': return new Division(); //pri=4; break;
      case '^': return new Power(); //pri=6; break;
      default:
        throw new UnsupportedOperationException("Operator "+(char)i);
    }
    //return new BinaryOperation(i, pri);
  }
  /**
   * Added for javacc generated parser.
   */
  public Expression createBinaryOperation(Expression firstOp,
      Expression secondOp, int i)
  {
    switch(i)
    {
      case '+': return new Addition(firstOp, secondOp); 
      case '*': return new Multiplication(firstOp, secondOp); 
      case '-': return new Subtraction(firstOp, secondOp); 
      case '/': return new Division(firstOp, secondOp); 
      case '^': return new Power(firstOp, secondOp); 
      default:
        throw new UnsupportedOperationException("Operator "+(char)i);
    }
  }
  /**
   * Creates and returns a new {@link BinaryOperation} with the type '*' and a
   * default priority.
   *
   * @see BinaryOperation#getType()
   * @see Expression#getPriority()
   * @return a new {@link BinaryOperation} with the type '*' and a
   * default priority.
   */
  public Expression createImplicitMultOperation()
  {
    return new Multiplication();//BinaryOperation('*', 4);
  }
  /**
   * Either creates and returns a new {@link FunctionCall} by calling
   * the method {@link #createFunctionCall(String)}
   * or calls {@link #createImplicitMultOperation()}
   * to return a new {@link BinaryOperation} with the type '*'.
   * The {@link DefinitionModel} will be searched for defined
   * {@link Function Functions} to decide
   * between function call or implicit multiplication.
   *
   * @return a new <code>FunctionCall</code> if <code>lastExp</code> is a
   * {@link Symbol} representing the name of a defined {@link Function},
   * otherwise it returns a new <code>BinaryOperation</code> of type '*'.
   * @see Symbol#toString()
   * @see DefinitionModel#containsFunction(String)
   * @param lastExp the left operand of the implicit operation.
   */
  public Expression createImplicitOperation(Expression lastExp)
  {
    if(lastExp instanceof Symbol)
    {
      String str=((Symbol)lastExp).toString();
      if(defModel.containsFunction(str))
      {
        return createFunctionCall(str);
      }
    }
    return createImplicitMultOperation();
  }
  /**
   * Added for javacc generated parser.
   * Checks if expr represents a Function. Therefor expr must be either
   * a Symbol or a UnaryOperation which operand is a Symbol. That Symbol
   * itself must be known as a function in the DefinitionModel.
   * 
   * This method returns the function name if expr is representing a function.
   * Otherwise, returns null.
   */
  public String getFunctionName(Expression expr)
  {
    Expression e;
    if(expr instanceof UnaryOperation)
      e=((UnaryOperation)expr).getOperand();
    else
      e=expr;
    if(e instanceof Symbol)
    {
      String str=((Symbol)e).toString();
      if(defModel.containsFunction(str))
      {
        return str;
      }
    }
    return null;
  }
  /**
   * Creates and returns a new {@link UnaryOperation} with the type
   * <code>i</code> and a default priority.
   * 
   * @throws UnsupportedOperationException if <code>i</code> is no ASCII-Code
   * for '+' or '-'.
   * @see UnaryOperation#getType()
   * @see Expression#getPriority()
   * @param i the ASCI-Code for the several <code>UnaryOperation</code> types
   * ('+' or '-').
   * @return a new {@link UnaryOperation} with the type
   * <code>i</code> and a default priority.
   */
  public Expression createUnaryOperation(int i)
  {
    int pri;
    switch(i)
    {
      case '+': return new PositiveSign();//pri=5; break;
      case '-': return new NegativeSign();//pri=5; break;
      default:
        throw new UnsupportedOperationException(
            "Operator "+(char)i+" is not possible in a unary context");
    }
    //return new UnaryOperation(i, pri);
  }
  /**
   * Added for javacc generated parser.
   */
  public Expression createUnaryOperation(Expression operand, int i)
  {
    switch(i)
    {
      case '+': return new PositiveSign(operand);
      case '-': return new NegativeSign(operand);
      default:
        throw new UnsupportedOperationException(
            "Operator "+(char)i+" is not possible in a unary context");
    }
  }
  /**
   * Creates and returns a new {@link FunctionCall} with the specified
   * <code>String</code> as the name of the {@link Function} which is calling
   * and a default priority.
   *
   * @return a new {@link FunctionCall} with a default priority.
   * @see Expression#getPriority()
   * @param functionName the name of the <code>Function</code> which is
   * calling.
   */
  public Expression createFunctionCall(String functionName)
  {
    return new FunctionCall(defModel.getFunction(functionName));
  }
  /**
   * Added for javacc generated parser.
   */
  public Expression createFunctionCall(String functionName, List param)
  {
    Function function = defModel.getFunction(functionName);
    if(function==null)
    	throw new UnknownDefinitionException("function \""+functionName+'"');
	return new FunctionCall(function, param);
  }  
  /**
   * Creates and returns a new {@link Symbol} representing a {@link Variable}
   * with the name <code>s</code>. Note that this <code>Symbol</code> will
   * maybe converted into a {@link FunctionCall} if an implicit operation is
   * detected at the next step of parsing.
   *
   * @see #createImplicitOperation(Expression)
   * @return a new <code>Symbol</code> representing a <code>Variable</code>
   * with the name <code>s</code>.
   * @param s the name of the <code>Variable</code> the created
   * <code>Symbol</code> will be representing.
   */
  public Expression createSymbol(String s)
  {
    return new Symbol(s);
  }
  /**
   * Creates and returns a new {@link Symbol} representing a constant number
   * with the value <code>d</code>.
   *
   * @return a new <code>Symbol</code> representing a constant number
   * with the value <code>d</code>.
   * @param d the constant value the created
   * <code>Symbol</code> will be representing.
   */
  public Expression createSymbol(double d)
  {
    return new Symbol(d);
  }
  /**
   * Added for javacc generated parser (and only because of consistency).
   */
  public Expression createList()
  {
    return new List();
  }
}
