/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2003-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.mathExpr.parser;

import java.util.Collection;
import java.util.Vector;

import de.jtem.mathExpr.ListFormChecker;
import de.jtem.mathExpr.WrongListFormException;
import de.jtem.mathExpr.evaluator.Context;
import de.jtem.mathExpr.evaluator.Evaluator;
import de.jtem.mathExpr.evaluator.Type;
//for doc-links:
import de.jtem.mathExpr.evaluator.Variable;
import de.jtem.mathExpr.evaluator.complexEvaluator.ComplexType;
import de.jtem.mathExpr.evaluator.realEvaluator.Real;
import de.jtem.mathExpr.evaluator.realEvaluator.RealType;
import de.jtem.mathExpr.parser.javacc.Parser;

/**
 * A <code>List</code> is an <code>Expression</code> which represents a list of
 * other <code>Expressions</code>.
 * For example, such <code>Lists</code> can represent mathematical vectors or
 * matrices.<br>
 * <br>
 * To check if a <code>List</code> represents the expected form you should use 
 * the class {@link ListFormChecker} which provides certain static methods.<br>
 * <br>
 * A <code>List</code> will be createt by a <code>Parser</code> whenever the
 * delimiter ',' will be found in the <code>String</code> to be parsed, e.g.:
 * {3, 4, 5} or f(3, 4, 5). See documentation of {@link Parser} for details.
 *
 * @author marcel
 */
public class List extends Expression
{
  /**
   * The <code>Vector</code> holding all entries of this <code>List</code>.
   * All entries are {@link Expression} objects.
   */
  Vector entries;
  /**
   * Creates a new <code>List</code> with 0 entries (an empty
   * <code>List</code>).
   */
  public List()
  {
    entries=new Vector(5);
  }
  /**
   * Adds the specified <code>Expression</code> at the end of this
   * <code>List</code>.
   * 
   * @param newEntry the <code>Expression</code> to be added as a new entry
   * of this <code>List</code>.
   */
  public void addEntry(Expression newEntry)
  {
    entries.add(newEntry);
  }
  /**
   * Returns the <code>Expression</code> which is located at the
   * <code>index'</code> position of this <code>List</code>.
   * 
   * @return the <code>Expression</code> which is located at the specified
   * <code>index</code> of this <code>List</code>.
   * @param index the position in this <code>List</code>.
   */
  public Expression getEntry(int index)
  {
    return (Expression)entries.get(index);
  }
  /**
   * Returns all <code>Expression</code> which are entries of this
   * <code>List</code>.
   * 
   * @return an <code>Expression</code> array of all entries of this
   * <code>List</code>.
   */
  public Expression[] getEntries()
  {
    int size=getEntryCount();
    Expression[] e=new Expression[size];
    for(int i=0; i<size; i++)
      e[i]=getEntry(i);
    return e;
  }
  /**
   * Returns the number of entries this <code>List</code> is containing.
   * 
   * @return the number of entries this <code>List</code> is containing.
   */
  public int getEntryCount()
  {
    return entries.size();
  }
  /**
   * Returns a constant value, because a <code>List</code> must not have a
   * priority. A <code>List</code> can only be operand of operations, like a
   * {@link Symbol}.
   * 
   * @return <code>-1</code>.
   */
  protected int getPriority()
  {
    return -1;
  }
  /**
   * Returns <code>true</code> if this <code>List</code> is empty or all entries
   * are constant, <code>false</code> otherwise.
   * 
   * @return <code>true</code> if this <code>List</code> is empty or all entries
   * are constant, <code>false</code> otherwise.
   */
  public boolean isConstant()
  {
    for(int i=0; i<getEntryCount(); i++)
      if(!((Expression)entries.get(i)).isConstant())
        return false;
    return true;
  }
  /**
   * Returns <code>true</code> if this <code>List</code> is empty or all entries
   * are complete, <code>false</code> otherwise.
   * 
   * @return <code>true</code> if this <code>List</code> is empty or all entries
   * are complete, <code>false</code> otherwise.
   */
  public boolean isComplete()
  {
    for(int i=0; i<getEntryCount(); i++)
      if(!getEntry(i).isComplete())
        return false;
    return true;
  }
  /**
   * Adds the names of the {@link Variable Variables} which are essential for
   * evaluation into the specified <code>Collection</code>. Therefor it gets the
   * essential <code>Variables</code> of all entries of this
   * <code>List</code>.
   * 
   * @param c the <code>Collection</code> the <code>Variable</code> names
   * will be added to.
   */
  public void getEssentialVariables(Collection c)
  {
    for(int i=0; i<getEntryCount(); i++)
      ((Expression)entries.get(i)).getEssentialVariables(c);
  }
  /**
   * Returns the <code>Evaluator</code> for this <code>List</code>.
   * <code>Type defaultType</code> specifies, how to deal with
   * <code>Symbols</code> in the expression-tree-hirarchy (evaluate any
   * <code>Symbol</code> as a {@link de.jtem.mfc.field.Complex} or a {@link Real} or
   * whatever?).
   * 
   * @return the <code>Evaluator</code> for this <code>List</code>.
   * @throws WrongListFormException if this <code>List</code> does not represent
   * a Vector or a Matrix.
   * @param defaultType the <code>Type</code> any <code>Symbol</code> will be
   * evaluated to.
   * @param context the <code>Context</code>, the <code>Evaluators<code> for
   * the <code>Symbols</code> are getting from.
   */
  public Evaluator getEvaluator(Type defaultType, Context context)
  {
    int size=getEntryCount();
    Evaluator[] e=new Evaluator[size];
    if(size==0)
      return defaultType.getEvaluator(this, e);
    if(ListFormChecker.representsVector(this))
    {
      boolean isComplex=false;
      for(int i=0; i<size; i++)
      {
        e[i]=getEntry(i).getEvaluator(defaultType, context);
        if(!isComplex && e[i].getReturnType()==ComplexType.TYPE)
          isComplex=true;
      }
      if(isComplex)//at least one entry is a Complex 
        return ComplexType.TYPE.getEvaluator(this, e);
      else//all entries are doubles (RealType.TYPE)
        return RealType.TYPE.getEvaluator(this, e);
    }
    //else if(ListFormChecker.representsMatrix(this))
    //  return ComplexMatrixType.TYPE.getEvaluator(this);
    else
      throw new WrongListFormException(this+": List does not represent a"+
          " vector or a matrix.");
  }
  /**
   * Returns a new List with this List's entries simplified.
   */
  public Expression simplify()
  {
    List simplifiedList=new List();
    for(int i=0; i<getEntryCount(); i++)
    {
      simplifiedList.addEntry(getEntry(i).simplify());
    }
    return simplifiedList;
  }
  /**
   * Returns a new List with this List entries differentiated.
   */
  public Expression differentiate(Symbol s)
  {
    List derivativeList=new List();
    for(int i=0; i<getEntryCount(); i++)
    {
      derivativeList.addEntry(getEntry(i).differentiate(s));
    }
    return derivativeList.simplify();
  }
  /**
   * Replaces all Symbols equal to s in this List
   * with the specified Expression. Therefor it replaces
   * all Symbols equal to s in this' entries. 
   * 
   * @see Symbol#replaceSymbol(Symbol,Expression)
   */
  public Expression replaceSymbol(Symbol s, Expression expr)
  {
    List newList=new List();
    for(int i=0; i<entries.size(); i++)
    {
      newList.addEntry(getEntry(i).replaceSymbol(s, expr));
    }
    return newList;
  }
  
  public Expression dissolveFunctionCalls()
  {
    List newList=new List();
    for(int i=0; i<entries.size(); i++)
      newList.addEntry(getEntry(i).dissolveFunctionCalls());
    return newList;    
  }
  /**
   * Returns <code>true</code> if the specified <code>Object</code> is also a
   * <code>List</code>, the number of entries is equal and all entries are equal to
   * this <code>List</code> entries, <code>false</code> otherwise.
   * 
   * @param o the object to compare this <code>List</code> against.
   * @return <code>true</code> if the specified <code>Object</code> is also a
   * <code>List</code>, the number of entries is equal and all entries are equal to
   * this <code>List</code> entries, <code>false</code> otherwise.
   * @see #getEntryCount()
   */
  public boolean equals(Object o)
  {
  	if(this==o) return true;
    if(!(o instanceof List))
      return false;
    List l=(List)o;
    if(getEntryCount()!=l.getEntryCount())
    	return false;
    for(int i=0; i<getEntryCount(); i++)
      if(!getEntry(i).equals(l.getEntry(i)))
        return false;
    return true;
  }
  /**
   * Returns a <code>String</code> representing this
   * <code>List</code>. It's now possible to parse the
   *  returned <code>String</code> by a {@link Parser} again.
   * 
   * @return a <code>String</code> representing this
   * <code>List</code>.
   */
  public String toString()
  {
    String text="{ ";
    for(int i=0; i<entries.size(); i++)
    {
      text+=entries.get(i);
      if(i!=entries.size()-1)
        text+=" , ";
    }
    text+=" }";
    return text;
  }
}
