/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2003-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.mathExpr.parser;

import de.jtem.mathExpr.PredefinedFunction;

/**
 * The BinaryOperation with the operator '^'. This class
 * implements the expression transformation methods 
 * {@link #simplify()} and {@link #differentiate(Symbol)} depending
 * on this certain Expression. 
 * 
 * @author marcel
 */
public class Power extends BinaryOperation
{
  /**
   * Creates a new Power object with no operands. See also constructor
   * {@link BinaryOperation#BinaryOperation(int, int)}.
   */
  public Power()
  {
    super('^', 6);
  }
  /**
   * Creates a new Power object with the specified operands. See also
   * constructor {@link
   * BinaryOperation#BinaryOperation(int, int, Expression, Expression)}.
   * 
   * @param op1
   * @param op2
   */
  public Power(Expression op1, Expression op2)
  {
    super('^', 6, op1, op2);
  }
  /**
   * Returns a simplified Expression of this Power. Therefor this
   * method calls simplify() on its operands. Then the following
   * possible Simplifications are:
   * <ul>
   * <li>a^0=1</li>
   * <li>0^a=0</li>
   * <li>a^1=a</li>
   * <li>1^a=1</li>
   * </ul>
   */
  public Expression simplify()
  {
    Expression base=operand1.simplify();
    Expression exponent=operand2.simplify();
    int minus=0;
    Expression baseWithoutSign;
    Expression exponentWithoutSign;
    if(base instanceof NegativeSign)
    {
      baseWithoutSign=((UnaryOperation)base).getOperand();
      minus=minus | 1;
    }
    else
      baseWithoutSign=base;
    if(exponent instanceof NegativeSign)
    {
      exponentWithoutSign=((UnaryOperation)exponent).getOperand();
      minus=minus | 2;
    }
    else
      exponentWithoutSign=exponent;
      
    if(exponentWithoutSign instanceof Symbol && exponentWithoutSign.isConstant())
    {
      Symbol s=(Symbol)exponentWithoutSign;
      if(s.getConstantValue()==0)
        return new Symbol(1);
      if(s.getConstantValue()==1)
        return (minus&2)!=0 ? new Power(base, exponent) : base;
    }
    if(baseWithoutSign instanceof Symbol && baseWithoutSign.isConstant())
    {
      Symbol s=(Symbol)baseWithoutSign;
      if(s.getConstantValue()==0)
        return baseWithoutSign;
      if(s.getConstantValue()==1 && (minus&1)==0)
        return base;
    }
    return new Power(base, exponent);
  }
  /**
   * Returns the derivative of this Power with respect to s.
   * In other words: Creates a new Multiplication in the following way
   * (f is this' first operand, g is this' second operand):<br>
   * (f^g)'=f^g * (g'*log(f) + f'*g/f)
   * <p>
   * If g is a constant <code>Symbol</code>, the following way is used:<br>
   * (f^g)'=g*(f^(g-1)) * f'
   * </p>
   * 
   * @see Symbol#isConstant()
   */
  public Expression differentiate(Symbol s)
  {
    if(operand2 instanceof Symbol && operand2.isConstant())
    {
      double newExponent=((Symbol)operand2).getConstantValue()-1;
      Expression pow=new Power(operand1, new Symbol(newExponent));
      Expression mult=new Multiplication(operand2, pow);
      return new Multiplication(mult, operand1.differentiate(s)).simplify();
    }
    else
    {
      Expression mult1=new Multiplication(operand2.differentiate(s),
          new FunctionCall(PredefinedFunction.LOG, FunctionCall.toList(operand1)));
      Expression mult2=new Multiplication(new Division(operand2, operand1),
          operand1.differentiate(s));
      return new Multiplication(this, new Addition(mult1, mult2)).simplify();
    }
    /*
     * If someday constant Expressions (e.g.: 3+4-5) will be simplified,
     * this algorithm could generally be better instead of using "if else":
     * (f^g)'=f^g * g' * log(f) + f^(g-1) * f' * g
    Expression mult1=new Multiplication(
        new Multiplication(this, operand2.differentiate(s)),
        new FunctionCall(PredefinedFunction.LOG, operand1));
    Expression mult2=new Multiplication(
        new Multiplication(operand2, operand1.differentiate(s)),
        new Power(operand1, new Subtraction(operand2, new Symbol(1))));
    return new Addition(mult1, mult2).simplify();
     */
  }
}
