/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2003-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.mathExpr.parser;

import java.util.Collection;

import de.jtem.mathExpr.UnknownDefinitionException;
import de.jtem.mathExpr.evaluator.Constant;
import de.jtem.mathExpr.evaluator.Context;
import de.jtem.mathExpr.evaluator.Evaluator;
import de.jtem.mathExpr.evaluator.Type;
//for doc-links:
import de.jtem.mathExpr.evaluator.Variable;
import de.jtem.mathExpr.evaluator.realEvaluator.Real;

/**
 * A <code>Symbol</code> is an <code>Expression</code> which can be a number
 * ({@link Constant}) or an identifier ({@link Variable}). The state must be set by
 * any constructor and can't be changed.<br>
 * <br>
 * <b>Attention:</b> If a <code>Symbol</code> represents an identifier
 * it must not be a <code>Variable</code> in any case. It can be also a
 * certain <code>static final Constant</code>, e.g.: {@link Constant#PI}
 * and {@link Constant#E}. 
 *
 * @see ExpressionFactory#createSymbol(double)
 * @see ExpressionFactory#createSymbol(String)
 */
public class Symbol extends Expression
{
  /**
   * The <code>String</code>, which represents the value if this
   * <code>Symbol</code> is constant, or the {@link Variable} name
   * if this <code>Symbol</code> is not constant.
   * 
   * @see #isConstant()
   */
  String str;
  /**
   * The double value this <code>Symbol</code> is representing if it's constant.
   * 
   * @see #isConstant()
   */
  private double constantValue;
  /**
   * Determines if this <code>Symbol</code> is constant or not.
   * 
   * @see #isConstant()
   */
  final boolean isConstant;
  /**
   * Creates a new <code>Symbol</code>, which represents a constant number
   * with the value <code>d</code>.
   * 
   * @see #getConstantValue()
   * @param d the constant double value this <code>Symbol</code> will be
   * representing.
   */
  public Symbol(double d)
  {
    str=String.valueOf(d);
    constantValue=d;
    isConstant=true;
  }
  /**
   * Creates a new <code>Symbol</code>, which represents a {@link Variable}
   * with the name <code>s</code>.
   * 
   * @param s the name of the <code>Variable</code> this <code>Symbol</code>
   * will be representing.
   */
  public Symbol(String s)
  {
    str=s;
    isConstant=false;
  }
  /**
   * Returns the constant double value this <code>Symbol</code> is representing
   * if this <code>Symbol</code> is constant. If this <code>Symbol</code>
   * represents a {@link Variable}, this method throws an
   * <code>IllegalStateException</code>.
   *
   * @see #isConstant()
   * @throws IllegalStateException if this <code>Symbol</code> represents a
   * <code>Variable</code>.
   * @return the constant double value this <code>Symbol</code> is representing
   * if this <code>Symbol</code> is constant.
   */
  public double getConstantValue()
  {
	  if(isConstant)
	    return constantValue;
	  else
	    throw new IllegalStateException("Not a constant value: "+str);
  }
  /**
   * Returns a constant value, because a <code>Symbol</code> must not have a
   * priority. A <code>Symbol</code> can only be operand of operations which
   * have priorities.
   * 
   * @return <code>-1</code>.
   */
  protected int getPriority()
  {
	  return -1;
  }
  /**
   * Returns <code>true</code> if this <code>Symbol</code> represents a constant
   * double value, <code>false</code> if this <code>Symbol</code> represents an
   * identifier of a {@link Variable} or of a static final {@link Constant}.
   * 
   * @see #getConstantValue()
   * @return <code>true</code> if this <code>Symbol</code> represents a constant
   * double value, <code>false</code> if this <code>Symbol</code> represents an
   * identifier of a {@link Variable} or of a static final {@link Constant}.
   */
  public boolean isConstant()//TODO: bad identifier???
  {
    return isConstant;
  }
  /**
   * Returns <code>true</code> whenever this method is called. As soon as a
   * <code>Symbol</code> was instantiated (as a number or an identifier), it is
   * ready to get evaluated. It doesn't matter if the appropriate {@link
   * Variable} is defined or not. Thant is the task for the
   * <code>Evaluator</code>. 
   * 
   * @see Context#getEvaluator(Symbol)
   * @see UnknownDefinitionException
   * @return <code>true</code>.
   */
  public boolean isComplete()
  {
    return true;
  }
  /**
   * Adds the name of the {@link Variable} represented by this
   * <code>Symbol</code> to the specified <code>Collection</code>.
   * 
   * @param c the <code>Collection</code> the <code>Variable</code> name
   * will be added to.
   */
  public void getEssentialVariables(Collection c)
  {
    if(!isConstant())
      c.add(str);
  }
  /**
   * Returns the <code>Evaluator</code> for this <code>Symbol</code>.
   * <code>Type defaultType</code> specifies, how to deal with this
   * <code>Symbols</code> (evaluate it as a {@link de.jtem.mfc.field.Complex} or a
   * {@link Real} or whatever?).
   * 
   * @return the <code>Evaluator</code> for this <code>Symbol</code>.
   * @param defaultType the <code>Type</code> any <code>Symbol</code> will be
   * evaluated to.
   * @param context the <code>Context</code>, the <code>Evaluator<code> for
   * this <code>Symbol</code> is getting from (in cases of representing a variable
   * and not a constant).
   */
  public Evaluator getEvaluator(Type defaultType, Context context)
  {
    return defaultType.getEvaluator(this, context);
  }
  /**
   * Returns this Symbol because a Symbol itself can't be simplified.
   */
  public Expression simplify()
  {
    return this;
  }
  /**
   * Returns a Symbol with constant value 1 if this Symbol equals
   * s. Otherwise returns a Symbol with constant value 0.
   */
  public Expression differentiate(Symbol s)
  {
    if(this.equals(s))
      return new Symbol(1);
    else
      return new Symbol(0);
  }
  /**
   * Returns the specified expression, if this Symbol equals s. Other
   * returns this.
   */
  public Expression replaceSymbol(Symbol s, Expression expr)
  {
    if(this.equals(s))
      return expr;
    else
      return this;
  }
  /**
   * Returns this.
   */
  public Expression dissolveFunctionCalls()
  {
    return this;
  }
  /**
   * Returns a hash code for this <code>Symbol</code>. It's the hash code from
   * the <code>String</code> representing the constant number respectively
   * the variable name.
   * 
   * @see String#hashCode()
   * @return a hash code for this <code>Symbol</code>.
   */
  public int hashCode()
  {
    return str.hashCode();
  }
  /**
   * Returns <code>true</code> if the specified <code>Object</code> is a
   * <code>Symbol</code> representing the same constant number respectively
   * the same variable name, <code>false</code> otherwise.
   * 
   * @param o the object to compare this <code>Symbol</code> against.
   * @return <code>true</code> if the specified <code>Object</code> is a
   * <code>Symbol</code> representing the same constant number respectively
   * the same variable name, <code>false</code> otherwise.
   */
  public boolean equals(Object o)
  {
    if(o==this) return true;
    if(!(o instanceof Symbol))
      return false;
	  Symbol s=(Symbol)o;
	  return str.equals(s.str);
  }
  /**
   * Returns the constant number respectively the variable name represented by
   * this <code>Symbol</code>.
   * 
   * @return the constant number respectively the variable name represented by
   * this <code>Symbol</code>.
   */
  public String toString()
  {
    return str;
  }
}
